
'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '34.0.0rc1',
    'branch_nick': 'master',
    'revision_id': '5b11b2faf62100dd846195ae2317d2a5605a100f',
    'revno': 12735
}
